from .CannotLoadError import CannotLoadError
from .NoElevatorError import NoElevatorError
from .NoFloorError import NoFloorError
from .NoDifficultyError import NoDifficultyError
from .Person import Person
from .Elevator import Elevator
from .Building import Building
from .Difficulty import Difficulty, get_difficulty, MAXIMUM_DIFFICULTY, MINIMUM_DIFFICULTY
from .Game import Game
from .curses import game_loop



class Person(object):
  def __init__(self, name: str, destination_floor: int) -> None:
    self._name = name
    self._destination_floor = destination_floor
    self._tick_started_waiting = 0

  def get_name(self) -> str:
    return self._name

  def get_destination_floor(self) -> int:
    return self._destination_floor

  def get_tick_started_waiting(self) -> int:
    return self._tick_started_waiting
  
  def set_tick_started_waiting(self, tick_started_waiting: int) -> None:
    self._tick_started_waiting = tick_started_waiting



import string

from . import Person, Elevator, NoFloorError, NoElevatorError, CannotLoadError

class Building(object):
  def __init__(self, number_of_floors: int, number_of_elevators: int) -> None:
    self._number_of_floors = number_of_floors
    # This includes an unused floor 0 for ease of development. It's worth 64 bits of memory.
    self._people_on_floors: list[list[Person]] = [[] for i in range(self._number_of_floors + 1)]
    self._number_of_elevators = number_of_elevators
    self._elevators: list[Elevator] = [Elevator(string.ascii_uppercase[i]) for i in range(self._number_of_elevators)]

  def send_elevator_to_floor(self, elevator_name: str, floor_number: int) -> None:
    if floor_number < 1 or floor_number > self._number_of_floors + 1:
      raise NoFloorError('Floor does not exist')
    for elevator in self._elevators:
      if elevator.get_name() == elevator_name:
        elevator.go_to_floor(floor_number)
        return
    raise NoElevatorError('There is no elevator by that name')

  def get_elevators(self) -> list[Elevator]:
    return self._elevators

  def get_people_on_floors(self) -> list[list[Person]]:
    return self._people_on_floors

  def open_door(self, elevator_name: str) -> None:
    for elevator in self._elevators:
      if elevator.get_name() == elevator_name:
        elevator.open_door()
        return
    raise NoElevatorError('There is no elevator by that name')

  def add_person_to_floor(self, person: Person, floor: int) -> None:
    if floor < 1 or floor > self._number_of_floors:
      raise NoFloorError("That floor does not exist")
    self._people_on_floors[floor].append(person)

  def load_passenger(self, person: str) -> None:
    for floor in range(1, self._number_of_floors + 1):
      for person_on_floor in self._people_on_floors[floor]:
        if person_on_floor.get_name() == person:
          for elevator in self._elevators:
            if elevator.get_current_floor() == floor and elevator.get_is_door_open():
              elevator.load_passenger(person_on_floor)
              self._people_on_floors[floor].remove(person_on_floor)
              return
    raise CannotLoadError('There is no person by that name, or no elevator on that floor with the door open')

  def unload_passenger(self, person: str) -> None:
    for elevator in self._elevators:
      for passenger in elevator.get_passengers():
        if passenger.get_name() == person:
          ex_passenger = elevator.unload_passenger(person)
          self._people_on_floors[elevator.get_current_floor()].append(ex_passenger)
          return
    raise CannotLoadError('There is no passenger by that name')

  def tick(self) -> None:
    """
    This doesn't include unloading passengers.
    That's left to the game to do later so it can get a list of persons unloaded successfully
    """
    for elevator in self._elevators:
      elevator.tick()

  def unload_passengers_who_have_arrived(self) -> list[Person]:
    total_arrived: list[Person] = []
    for elevator in self._elevators:
      arrived = elevator.unload_passengers_who_have_arrived()
      total_arrived.extend(arrived)
    return total_arrived



from . import Person

class Elevator(object):
  def __init__(self, name: str) -> None:
    self._name = name
    self._current_floor = 1
    self._destination_floor = 1
    self._is_door_open = False
    self._passengers: list[Person] = []

  def get_name(self) -> str:
    return self._name

  def get_current_floor(self) -> int:
    return self._current_floor

  def get_destination_floor(self) -> int:
    return self._destination_floor

  def get_is_door_open(self) -> bool:
    return self._is_door_open

  def get_passengers(self) -> list[Person]:
    return self._passengers

  def go_to_floor(self, destination_floor: int) -> None:
    self._is_door_open = False
    self._destination_floor = destination_floor

  def open_door(self) -> None:
    self._destination_floor = self._current_floor
    self._is_door_open = True

  def load_passenger(self, passenger: Person) -> None:
    self._passengers.append(passenger)

  def unload_passenger(self, passenger_name: str) -> Person:
    if not self._is_door_open:
      raise Exception('Elevator door is not open')
    for passenger in self._passengers:
      if passenger.get_name() == passenger_name:
        self._passengers.remove(passenger)
        return passenger
    raise IndexError('There is no passenger by that name on this elevator')

  def unload_passengers_who_have_arrived(self) -> list[Person]:
    arrived = []
    for passenger in self._passengers:
      if self._is_door_open and self._current_floor == passenger.get_destination_floor():
        arrived.append(passenger)
    for passenger in arrived:
      self._passengers.remove(passenger)
    return arrived

  def tick(self) -> None:
    if self._destination_floor > self._current_floor:
      self._current_floor += 1
    elif self._destination_floor < self._current_floor:
      self._current_floor -= 1



import string
import random
from time import time
from typing import Any

from . import Building, Person, Elevator, get_difficulty

_SCORE_FEEL_GOOD_MULTIPLIER = 100

def _subarray_count(array: list[list[Any]]):
  count = 0
  for arr in array:
    count += len(arr)
  return count

class Game(object):
  def __init__(self, difficulty: int) -> None:
    self.score = 0.0
    self.is_game_over = False
    self.ticks_passed = -1
    self.__difficulty = get_difficulty(difficulty)
    self.people_to_arrive = self.__difficulty.number_of_people
    self.number_of_floors = self.__difficulty.number_of_floors
    self.__time_of_last_tick = 0
    self.__building = Building(self.number_of_floors, self.__difficulty.number_of_elevators)

  def tick_time_check(self) -> bool:
    now = time()
    if now - self.__time_of_last_tick > self.__difficulty.seconds_between_ticks:
      self.ticks_passed += 1
      self.__time_of_last_tick = now
      self._tick()
      return True
    return False

  def _tick(self) -> None:
    self.__building.tick()
    arrived_people = self.__building.unload_passengers_who_have_arrived()
    for person in arrived_people:
      self.score += _SCORE_FEEL_GOOD_MULTIPLIER * self.number_of_floors * self.__difficulty.difficulty / (self.ticks_passed - person.get_tick_started_waiting())
    if self.ticks_passed % self.__difficulty.ticks_between_people == 0 and self.people_to_arrive > 0:
      destination_floor = random.randint(1, self.number_of_floors)
      possible_names = set(string.ascii_lowercase)
      all_people_in_building: list[Person] = []
      for floor in self.__building.get_people_on_floors():
        all_people_in_building.extend(floor)
      for elevator in self.get_elevators():
        all_people_in_building.extend(elevator.get_passengers())
      names_in_use = set([person.get_name() for person in all_people_in_building])
      available_names = list(possible_names - names_in_use)
      available_names.sort()
      if len(available_names) > 0:
        name = list(available_names)[0]
        person = Person(name, destination_floor)
        source_floor = person.get_destination_floor()
        while source_floor == person.get_destination_floor():
          source_floor = random.randint(1, self.number_of_floors)
        person.set_tick_started_waiting(self.ticks_passed)
        self.__building.add_person_to_floor(person, source_floor)
        self.people_to_arrive -= 1

    # Game-state change
    count_people_on_floors = _subarray_count(self.__building.get_people_on_floors())
    count_people_on_elevators = _subarray_count([elevator.get_passengers() for elevator in self.__building.get_elevators()])
    if self.people_to_arrive < 1 and count_people_on_floors < 1 and count_people_on_elevators < 1:
      self.is_game_over = True

  def get_elevators(self) -> list[Elevator]:
    return self.__building.get_elevators()

  def get_people_on_floors(self) -> list[list[Person]]:
    return self.__building.get_people_on_floors()

  def send_elevator_to_floor(self, elevator_name: str, floor_number: int) -> None:
    self.__building.send_elevator_to_floor(elevator_name, floor_number)
  
  def open_door(self, elevator_name: str) -> None:
    self.__building.open_door(elevator_name)

  def load_or_unload_passenger(self, person_name: str) -> None:
    for floor in self.__building.get_people_on_floors():
      for person in floor:
        if person.get_name() == person_name:
          self.__building.load_passenger(person_name)
          return
    self.__building.unload_passenger(person_name)



from curses import wrapper, window as Window, curs_set, A_BOLD, A_DIM, COLOR_BLUE, COLOR_WHITE, init_pair, color_pair, newpad
from time import sleep
from typing import Union
import string

from . import Game, MAXIMUM_DIFFICULTY, MINIMUM_DIFFICULTY

game: Game = Game(MINIMUM_DIFFICULTY)
selected_elevator: Union[str, None] = None
auto_load = True

def game_loop():
  global game
  while True:
    difficulty = ''
    try:
      difficulty = input('Choose a difficulty({} - {}) or "exit" to quit: '.format(MINIMUM_DIFFICULTY, MAXIMUM_DIFFICULTY))
    except KeyboardInterrupt:
      print('')
      break
    if difficulty == 'exit':
      break
    difficulty_int = -1
    try:
      difficulty_int = int(difficulty)
    except:
      pass
    if difficulty_int < MINIMUM_DIFFICULTY or difficulty_int > MAXIMUM_DIFFICULTY:
      print('Invalid difficulty')
      continue

    game = Game(difficulty_int)
    try:
      wrapper(__play_game)
    except KeyboardInterrupt:
      pass
    print('You scored {:,.0f} on difficulty {}'.format(game.score, difficulty))

def __play_game(window: Window):
  global game
  global selected_elevator
  global auto_load
  init_pair(1, COLOR_WHITE, COLOR_BLUE)
  try:
    curs_set(False)
  except:
    pass
  window.nodelay(True)
  while not game.is_game_over:
    if game.tick_time_check():
      render(game, window)
      if auto_load:
        people_on_floors = game.get_people_on_floors()
        for elevator in game.get_elevators():
          if elevator.get_is_door_open():
            for person in people_on_floors[elevator.get_current_floor()]:
              game.load_or_unload_passenger(person.get_name())
    try:
      input = window.getkey()
      render(game, window)
      if input in string.ascii_lowercase:
        game.load_or_unload_passenger(input) # TODO catch error
      elif input in string.ascii_uppercase:
        selected_elevator = input
      elif input in string.digits:
        if selected_elevator is not None:
          game.send_elevator_to_floor(selected_elevator, int(input)) # TODO catch error
      elif input == '!':
        if selected_elevator is not None:
          game.open_door(selected_elevator) # TODO catch error
    except:
      pass
    sleep(1 / 24)

def render(game: Game, window: Window):
  global selected_elevator
  global auto_load
  
  window.clear()

  # Status bar
  window.move(1, 1)
  window.addstr('Time Passed ')
  window.addstr('{:3,}'.format(game.ticks_passed), A_BOLD)
  window.addstr(' | Score ')
  window.addstr('{:6,.0f}'.format(game.score), A_BOLD)
  window.addstr(' | People yet to arrive ')
  window.addstr('{:2,}'.format(game.people_to_arrive), A_BOLD)
  
  # Messages go to 2, 1
  
  # Elevator Header
  floor_start_row = 4
  window.move(floor_start_row, 1)
  window.addstr('Floor', A_DIM)
  for elevator in game.get_elevators():
    window.addstr(' ')
    if elevator.get_is_door_open():
      window.addstr('{}'.format(elevator.get_name()), color_pair(1))
    else:
      window.addstr('{}'.format(elevator.get_name()))
  
  # Floors
  floor_current_row = floor_start_row 
  for floor in range(game.number_of_floors, 0, -1):
    floor_current_row += 1
    window.move(floor_current_row, 1)
    window.addstr('{}'.format(floor))
    for elevator_index in range(len(game.get_elevators())):
      elevator = game.get_elevators()[elevator_index]
      if elevator.get_current_floor() == floor:
        window.move(floor_current_row, 6 + 2 * elevator_index)
        window.addstr('{:2.0f}'.format(len(elevator.get_passengers())))
  
  # People header
  person_start_column = 32
  person_current_row = floor_start_row
  window.move(person_current_row, person_start_column)
  window.addstr('Name Location Destination', A_DIM)
  
  LINES_BELOW_PEOPLE = 5
  
  # People on elevators
  for elevator in game.get_elevators():
    for person in elevator.get_passengers():
      if person_current_row + LINES_BELOW_PEOPLE > window.getmaxyx()[0]:
        break
      person_current_row += 1
      window.move(person_current_row, person_start_column)
      window.addstr('{}    {}({})     {}'.format(person.get_name(), elevator.get_name(), elevator.get_current_floor(), person.get_destination_floor()))

  # People on floorsA
  window.move(2, 1)
  for floor_index in range(1, game.number_of_floors + 1):
    floor = game.get_people_on_floors()[floor_index]
    for person in floor:
      if person_current_row + LINES_BELOW_PEOPLE > window.getmaxyx()[0]:
        break
      person_current_row += 1
      window.move(person_current_row, person_start_column)
      window.addstr('{}    {}        {}'.format(person.get_name(), floor_index, person.get_destination_floor()))

  # Commands
  commands_row = max(person_current_row, floor_current_row) + 2
  window.move(commands_row, 1)
  if selected_elevator is None:
    window.addstr('Uppercase letter to select an elevator', A_DIM)
  else:
    window.addstr('Elevator ', A_DIM)
    window.addstr('{}'.format(selected_elevator))
    window.addstr(' selected | Number to send elevator to floor | ! to open door', A_DIM)

  # Auto loading
  window.move(commands_row + 1, 1)
  if auto_load:
    window.addstr('Auto-load | @ to disable auto-loading elevators', A_DIM)
  else:
    window.addstr('@ to enable auto-loading elevators | Lowercase to manually load/unload person', A_DIM)

  window.move(3, 1)



from . import NoDifficultyError

MINIMUM_DIFFICULTY = 1
MAXIMUM_DIFFICULTY = 10

class Difficulty(object):
  def __init__(self, difficulty: int, number_of_people: int, ticks_between_people: int, seconds_between_ticks: float, number_of_floors: int, number_of_elevators: int) -> None:
    self.difficulty = difficulty
    self.number_of_people = number_of_people
    self.ticks_between_people = ticks_between_people
    self.number_of_floors = number_of_floors
    self.number_of_elevators = number_of_elevators
    self.seconds_between_ticks = seconds_between_ticks

def get_difficulty(difficulty: int) -> Difficulty:
  if difficulty == 1:
    return Difficulty(difficulty, 20, 6, 3.0, 4, 1)
  if difficulty == 2:
    return Difficulty(difficulty, 20, 7, 2.2, 5, 1)
  if difficulty == 3:
    return Difficulty(difficulty, 20, 7, 2.0, 7, 2)
  if difficulty == 4:
    return Difficulty(difficulty, 20, 6, 1.8, 8, 3)
  if difficulty == 5:
    return Difficulty(difficulty, 20, 6, 1.5, 9, 4)
  if difficulty == 6:
    return Difficulty(difficulty, 20, 5, 1.2, 9, 4)
  if difficulty == 7:
    return Difficulty(difficulty, 26, 4, 1.2, 9, 4)
  if difficulty == 8:
    return Difficulty(difficulty, 26, 4, 1.1, 9, 5)
  if difficulty == 9:
    return Difficulty(difficulty, 26, 4, 1.0, 9, 5)
  if difficulty == 10:
    return Difficulty(difficulty, 26, 3, 1.0, 9, 5)
  raise NoDifficultyError('That difficulty does not exist')
